<?php

declare(strict_types=1);

namespace DigiComp\PhpCodesniffer\DigiComp\Sniffs\Annotations;

use DigiComp\PhpCodesniffer\DigiComp\Sniffs\MemberVarUtility;
use PHP_CodeSniffer\Files\File;
use PHP_CodeSniffer\Sniffs\Sniff;
use PHP_CodeSniffer\Util\Tokens;

class ReturnIsNotVoidSniff implements Sniff
{
    public function register()
    {
        return [\T_FUNCTION];
    }

    public function process(File $phpcsFile, $stackPtr)
    {
        $tokens = $phpcsFile->getTokens();
        $ignore = Tokens::$methodPrefixes;
        $ignore[\T_WHITESPACE] = \T_WHITESPACE;

        $commentEnd = MemberVarUtility::findCommentEndWithoutAttributes($tokens, $ignore, $stackPtr);

        if ($tokens[$commentEnd]['code'] === \T_COMMENT) {
            // Inline comments might just be closing comments for
            // control structures or functions instead of function comments
            // using the wrong comment type. If there is other code on the line,
            // assume they relate to that code.
            $prev = $phpcsFile->findPrevious($ignore, ($commentEnd - 1), null, true);
            if ($prev !== false && $tokens[$prev]['line'] === $tokens[$commentEnd]['line']) {
                $commentEnd = $prev;
            }
        }

        if ($tokens[$commentEnd]['code'] !== \T_DOC_COMMENT_CLOSE_TAG) {
            return;
        }

        $commentStart = $tokens[$commentEnd]['comment_opener'];

        foreach ($tokens[$commentStart]['comment_tags'] as $tag) {
            if (($tokens[$tag]['content'] === '@return') && 'void' === $tokens[$tag + 2]['content']) {
                $fix = $phpcsFile->addFixableError('Return type void must not be annotated.', $tag, 'WrongReturnType');
                if ($fix) {
                    $nextTokenIndex = \array_search($tag, $tokens[$commentStart]['comment_tags'], true) + 1;
                    if (!isset($tokens[$commentStart]['comment_tags'][$nextTokenIndex])) {
                        $tag -= 2;
                        $nextToken = $commentEnd;
                    } else {
                        $nextToken = $tokens[$commentStart]['comment_tags'][$nextTokenIndex];
                    }
                    $phpcsFile->fixer->beginChangeset();
                    for ($tokenToDelete = $tag; $tokenToDelete < $nextToken; $tokenToDelete++) {
                        $phpcsFile->fixer->replaceToken($tokenToDelete, '');
                    }
                    $phpcsFile->fixer->endChangeset();
                }
            }
        }
    }
}
